#!/usr/bin/env python

"""
Generate a dummy randomization list.

This trial is randomized by site so all assignments are
the same within a site. Use this util to generate a dummy
randomization_list.csv for import into the RandomizationList
model. Patient registration always refers to and updates the
RandomizationList model.

"""

import csv
from typing import Optional

from edc_sites import get_site_id


def generate_randomization_list(
    all_sites=None,
    country=None,
    site_name=None,
    assignment: Optional[list] = None,
    slots: Optional[int] = None,
    write_header: Optional[bool] = None,
    filename=None,
    assignment_map=None,
):
    """
    Adds slots to  a dummy `randomisation` list file where all assignments are the same
    for each slot.
    """
    slots = slots or 10
    assignment_map = assignment_map or ["intervention", "control"]
    if assignment not in assignment_map:
        raise ValueError(f"Invalid assignment. Got {assignment}")

    # get site ID and write the file
    site_id = get_site_id(site_name, sites=all_sites[country])
    with open(filename, "a+", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=["sid", "assignment", "site_name", "country"])
        if write_header:
            writer.writeheader()
        for j in range(1, int(slots)):
            sid = str(j).zfill(len(str(slots)))
            writer.writerow(
                dict(
                    sid=f"{site_id}{sid}",
                    assignment=assignment,
                    site_name=site_name,
                    country=country,
                )
            )

    print(f"(*) Added {slots} slots for {site_name}.")



import csv
import os
import random
from tempfile import mkdtemp

from ..constants import ACTIVE, PLACEBO

default_assignments = [ACTIVE, PLACEBO]


def make_test_list(
    full_path=None,
    assignments=None,
    site_names=None,
    count=None,
    first_sid=None,
    per_site=None,
):
    first_sid = first_sid or 0
    if per_site:
        site_names = site_names * per_site
        count = len(site_names)
        gen_site_name = (x for x in site_names)
    else:
        count = count or 50
        gen_site_name = (random.choice(site_names) for i in range(0, 50))

    if not full_path:
        full_path = os.path.join(mkdtemp(), "randomizationlist.csv")
    assignments = assignments or default_assignments
    with open(full_path, "w") as f:
        writer = csv.DictWriter(f, fieldnames=["sid", "assignment", "site_name"])
        writer.writeheader()
        n = 0
        for i in range(first_sid, count + first_sid):
            n += 1
            assignment = random.choice(assignments)
            writer.writerow(dict(sid=i, assignment=assignment, site_name=next(gen_site_name)))
    return full_path



from edc_auth.site_auths import site_auths

from .auth_objects import (
    RANDO_BLINDED,
    RANDO_UNBLINDED,
    get_rando_permissions_codenames,
    make_randomizationlist_view_only,
    update_rando_group_permissions,
)

site_auths.add_group(get_rando_permissions_codenames, name=RANDO_BLINDED, view_only=True)
site_auths.add_group(get_rando_permissions_codenames, name=RANDO_UNBLINDED, view_only=True)
site_auths.add_post_update_func(update_rando_group_permissions)
site_auths.add_post_update_func(make_randomizationlist_view_only)



import os
import sys
from warnings import warn

from django.apps import AppConfig as DjangoAppConfig
from django.conf import settings
from django.core.checks.registry import register

from .site_randomizers import site_randomizers
from .system_checks import randomization_list_check


class AppConfig(DjangoAppConfig):
    name = "edc_randomization"
    verbose_name = "Edc Randomization"
    has_exportable_data = True
    include_in_administration_section = True

    def ready(self):
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_randomizers.autodiscover()
        register(randomization_list_check, deploy=True)
        sys.stdout.write(f" Done loading {self.verbose_name} ...\n")

    @property
    def randomization_list_path(self):
        warn(
            "Use of settings.RANDOMIZATION_LIST_PATH has been deprecated. "
            "See site_randomizers in edc_randomization"
        )
        return os.path.join(settings.RANDOMIZATION_LIST_PATH)



from django.conf import settings
from django.views.generic.base import TemplateView
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):

    template_name = f"edc_randomization/bootstrap{settings.EDC_BOOTSTRAP}/home.html"
    navbar_name = "edc_randomization"
    navbar_selected_item = "edc_randomization"



from edc_model import models as edc_models

from .model_mixin import RandomizationListModelMixin


class RandomizationList(RandomizationListModelMixin, edc_models.BaseUuidModel):
    class Meta(RandomizationListModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        pass



import csv
import os
import sys

from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import OperationalError, ProgrammingError

from edc_randomization.site_randomizers import site_randomizers


class RandomizationListError(Exception):
    pass


class RandomizationListVerifier:

    """Verifies the RandomizationList upon instantiation."""

    default_fieldnames = ["sid", "assignment", "site_name"]

    def __init__(self, randomizer_name=None, fieldnames=None):
        self.messages = []

        self.randomizer = site_randomizers.get(randomizer_name)
        if not self.randomizer:
            raise RandomizationListError(f"Randomizer not registered. Got `{randomizer_name}`")
        self.fieldnames = fieldnames or self.default_fieldnames
        try:
            self.count = self.randomizer.model_cls().objects.all().count()
        except (ProgrammingError, OperationalError) as e:
            self.messages.append(str(e))
        else:
            if self.count == 0:
                self.messages.append(
                    "Randomization list has not been loaded. "
                    "Run the 'import_randomization_list' management command "
                    "to load before using the system. "
                    "Resolve this issue before using the system."
                )

            else:
                if not self.randomizer.get_randomization_list_fullpath() or not os.path.exists(
                    self.randomizer.get_randomization_list_fullpath()
                ):
                    self.messages.append(
                        f"Randomization list file does not exist but SIDs "
                        f"have been loaded. Expected file "
                        f"{self.randomizer.get_randomization_list_fullpath()}. "
                        f"Resolve this issue before using the system."
                    )
                else:
                    message = self.verify_list()
                    if message:
                        self.messages.append(message)
        if self.messages:
            if (
                "migrate" not in sys.argv
                and "makemigrations" not in sys.argv
                and "import_randomization_list" not in sys.argv
            ):
                raise RandomizationListError(", ".join(self.messages))

    def verify_list(self):

        message = None

        with open(self.randomizer.get_randomization_list_fullpath(), "r") as f:
            reader = csv.DictReader(f)
            for index, row in enumerate(reader):
                row = {k: v.strip() for k, v in row.items() if k}
                if index == 0:
                    continue
                message = self.inspect_row(index, row)
                if message:
                    break
        if not message:
            if self.count != index + 1:
                message = (
                    f"Randomization list count is off. Expected {index + 1} (CSV). "
                    f"Got {self.count} (model_cls). See file "
                    f"{self.randomizer.get_randomization_list_fullpath()}. "
                    f"Resolve this issue before using the system."
                )
        return message

    def inspect_row(self, index, row):
        message = None
        obj1 = self.randomizer.model_cls().objects.all().order_by("sid")[index]
        try:
            obj2 = self.randomizer.model_cls().objects.get(sid=row["sid"])
        except ObjectDoesNotExist:
            message = f"Randomization file has an invalid SID. Got {row['sid']}"
        else:
            if obj1.sid != obj2.sid:
                message = (
                    f"Randomization list has invalid SIDs. List has invalid SIDs. "
                    f"File data does not match model data. See file "
                    f"{self.randomizer.get_randomization_list_fullpath()}. "
                    f"Resolve this issue before using the system. "
                    f"Problem started on line {index + 1}. "
                    f'Got \'{row["sid"]}\' != \'{obj1.sid}\'.'
                )
            if not message:
                assignment = self.randomizer.get_assignment(row)
                if obj2.assignment != assignment:
                    message = (
                        f"Randomization list does not match model. File data "
                        f"does not match model data. See file "
                        f"{self.randomizer.get_randomization_list_fullpath()}. "
                        f"Resolve this issue before using the system. "
                        f"Got '{assignment}' != '{obj2.assignment}' for sid={obj2.sid}."
                    )
                elif obj2.site_name != row["site_name"]:
                    message = (
                        f"Randomization list does not match model. File data "
                        f"does not match model data. See file "
                        f"{self.randomizer.get_randomization_list_fullpath()}. "
                        f"Resolve this issue before using the system. "
                        f'Got \'{obj2.site_name}\' != \'{row["site_name"]}\' '
                        f"for sid={obj2.sid}."
                    )
        return message



from django.core.management.base import BaseCommand, CommandError

from edc_randomization.site_randomizers import site_randomizers

from ...randomization_list_importer import (
    RandomizationListAlreadyImported,
    RandomizationListImporter,
    RandomizationListImportError,
)


class Command(BaseCommand):
    help = "Import randomization list"

    def add_arguments(self, parser):
        parser.add_argument(
            "--path",
            dest="path",
            default=None,
            help="full path to CSV file. Default: app_config.randomization_list_path",
        )

        parser.add_argument(
            "--name", dest="name", default="default", help="name of randomizer"
        )

        parser.add_argument(
            "--force-add",
            dest="add",
            default="NO",
            help="overwrite existing data. CANNOT BE UNDONE!!",
        )

        parser.add_argument(
            "--dryrun",
            dest="dryrun",
            default="NO",
            help="Dry run. No changes will be made",
        )

        parser.add_argument("--user", dest="user", default=None, help="user")

        parser.add_argument("--revision", dest="revision", default=None, help="revision")

    def handle(self, *args, **options):
        """Note: You may not need to do this.

        `import_list()` is usually called when the `randomizer` class
        is first instantiated.
        """
        add = True if options["add"] and options["add"].lower() == "yes" else False
        dryrun = True if options["dryrun"] and options["dryrun"].lower() == "yes" else False
        name = options["name"]
        username = options["user"]
        revision = options["revision"]
        randomizer_cls = site_randomizers.get(name)
        importer = RandomizationListImporter(
            randomizer_cls=randomizer_cls,
            add=add,
            dryrun=dryrun,
            username=username,
            revision=revision,
        )
        try:
            importer.import_list()
        except (
            RandomizationListImportError,
            RandomizationListAlreadyImported,
            FileNotFoundError,
        ) as e:
            raise CommandError(e)



from tempfile import mkdtemp

from ..randomizer import Randomizer

tmpdir = mkdtemp()


class MyRandomizer(Randomizer):
    name = "my_randomizer"
    model = "edc_randomization.myrandomizationlist"
    randomization_list_path = tmpdir


class MyOtherRandomizer(Randomizer):
    name = "my_other_randomizer"
    model = "edc_randomization.myotherrandomizationlist"
    randomization_list_path = tmpdir



# Generated by Django 2.2.6 on 2019-11-08 05:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_randomization", "0003_auto_20191024_0426")]

    operations = [
        migrations.AlterField(
            model_name="randomizationlist",
            name="allocated_site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.Site",
            ),
        )
    ]



import os

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from edc_registration.utils import get_registered_subject_model_cls

from edc_randomization.randomization_list_importer import (
    RandomizationListAlreadyImported,
    RandomizationListImporter,
)

from .constants import DEFAULT_ASSIGNMENT_MAP
from .randomization_list_verifier import RandomizationListVerifier

RANDOMIZED = "RANDOMIZED"

assignment_map = getattr(settings, "EDC_RANDOMIZATION_ASSIGNMENT_MAP", DEFAULT_ASSIGNMENT_MAP)


class RandomizationError(Exception):
    pass


class RandomizationListFileNotFound(Exception):
    pass


class RandomizationListNotLoaded(Exception):
    pass


class AlreadyRandomized(ValidationError):
    pass


class AllocationError(Exception):
    pass


class InvalidAllocation(Exception):
    pass


class InvalidAssignment(Exception):
    pass


class Randomizer:
    """Selects and uses the next available slot in model
    RandomizationList (cls.model) for this site. A slot is used
    when the subject identifier is not None.

    This is the default randomizer class and is registered with
    `site_randomizer` by default. To prevent registration set
    settings.EDC_RANDOMIZATION_REGISTER_DEFAULT_RANDOMIZER=False.
    """

    name = "default"
    model = "edc_randomization.randomizationlist"
    assignment_map = assignment_map
    filename = "randomization_list.csv"
    randomization_list_path = settings.EDC_RANDOMIZATION_LIST_PATH
    is_blinded_trial = True
    importer_cls = RandomizationListImporter

    def __init__(
        self, subject_identifier=None, report_datetime=None, site=None, user=None, **kwargs
    ):
        self._model_obj = None
        self._registered_subject = None
        self.subject_identifier = subject_identifier
        self.allocated_datetime = report_datetime
        self.site = site
        self.user = user
        if not os.path.exists(self.get_randomization_list_fullpath()):
            raise RandomizationListFileNotFound(
                "Randomization list file not found. "
                f"Got {self.get_randomization_list_fullpath()}. See {self}."
            )
        self.check_loaded()
        # force query, will raise if already randomized
        self.get_registered_subject()
        # will raise if already randomized
        self.randomize()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.name},{self.get_randomization_list_fullpath()})"
        )

    def __str__(self):
        return f"<{self.name} for file {self.get_randomization_list_fullpath()}>"

    def randomize(self):
        required_instance_attrs = dict(
            subject_identifier=self.subject_identifier,
            allocated_datetime=self.allocated_datetime,
            user=self.user,
            site=self.site,
            **self.extra_required_instance_attrs,
        )

        if not all(required_instance_attrs.values()):
            raise RandomizationError(
                f"Randomization failed. Insufficient data. Got {required_instance_attrs}."
            )
        self.model_obj.subject_identifier = self.subject_identifier
        self.model_obj.allocated_datetime = self.allocated_datetime
        self.model_obj.allocated_user = self.user
        self.model_obj.allocated_site = self.site
        self.model_obj.allocated = True
        self.model_obj.save()
        # requery
        self._model_obj = self.model_cls().objects.get(
            subject_identifier=self.subject_identifier,
            allocated=True,
            allocated_datetime=self.allocated_datetime,
        )
        self.registered_subject.sid = self.model_obj.sid
        self.registered_subject.randomization_datetime = self.model_obj.allocated_datetime
        self.registered_subject.registration_status = RANDOMIZED
        self.registered_subject.randomization_list_model = self.model_obj._meta.label_lower
        self.registered_subject.save()
        # requery
        self._registered_subject = get_registered_subject_model_cls().objects.get(
            subject_identifier=self.subject_identifier, sid=self.model_obj.sid
        )

    @property
    def extra_required_instance_attrs(self):
        """Returns a dict of extra attributes that must have
        value on self.
        """
        return {}

    @classmethod
    def model_cls(cls, apps=None):
        return (apps or django_apps).get_model(cls.model)

    @classmethod
    def get_randomization_list_fullpath(cls):
        return os.path.expanduser(os.path.join(cls.randomization_list_path, cls.filename))

    @classmethod
    def get_assignment(cls, row):
        """Returns assignment (text) after checking validity."""
        assignment = row["assignment"]
        if assignment not in cls.assignment_map:
            raise InvalidAssignment(
                f"Invalid assignment. Expected one of {list(cls.assignment_map.keys())}. "
                f"Got `{assignment}`. "
                f"See randomizer `{cls.name}` {repr(cls)}. "
            )
        return assignment

    @classmethod
    def get_allocation(cls, row):
        """Returns an integer allocation for the given
        assignment or raises.
        """
        assignment = cls.get_assignment(row)
        return cls.assignment_map.get(assignment)

    @property
    def sid(self):
        """Returns the SID."""
        return self.model_obj.sid

    @classmethod
    def check_loaded(cls):
        try:
            cls.import_list(overwrite=False)
        except RandomizationListAlreadyImported:
            pass
        # except RandomizationListImportError as e:
        #     sys.stdout.write(f"RandomizationListImportError. {e}\n")
        if cls.model_cls().objects.all().count() == 0:
            raise RandomizationListNotLoaded(
                "Randomization list has not been loaded. "
                "You may need to run the management command or check "
                "the path or format of the `randomization list` file. "
                f"See {repr(cls)}."
            )

    @property
    def model_obj(self):
        """Returns a "rando" model instance by selecting
        the next available SID.
        """
        if not self._model_obj:
            try:
                obj = self.model_cls().objects.get(subject_identifier=self.subject_identifier)
            except ObjectDoesNotExist:
                opts = dict(site_name=self.site.name, **self.extra_model_obj_options)
                self._model_obj = (
                    self.model_cls()
                    .objects.filter(subject_identifier__isnull=True, **opts)
                    .order_by("sid")
                    .first()
                )
                if not self._model_obj:
                    fld_str = ", ".join([f"{k}=`{v}`" for k, v in opts.items()])
                    raise AllocationError(
                        f"Randomization failed. No additional SIDs available for {fld_str}."
                    )
            else:
                raise AlreadyRandomized(
                    "Subject already randomized. "
                    f"Got {obj.subject_identifier} SID={obj.sid}. "
                    "Something is wrong. Are registered_subject and "
                    f"{self.model_cls()._meta.label_lower} out of sync?.",
                    code=self.model_cls()._meta.label_lower,
                )
        return self._model_obj

    @property
    def extra_model_obj_options(self):
        """Returns a dict of extra key/value pair for filtering the
        "rando" model.
        """
        return {}

    def get_registered_subject(self):
        return self.registered_subject

    @property
    def registered_subject(self):
        """Returns an instance of the registered subject model."""
        if not self._registered_subject:
            try:
                self._registered_subject = get_registered_subject_model_cls().objects.get(
                    subject_identifier=self.subject_identifier, sid__isnull=True
                )
            except ObjectDoesNotExist:
                try:
                    obj = get_registered_subject_model_cls().objects.get(
                        subject_identifier=self.subject_identifier
                    )
                except ObjectDoesNotExist:
                    raise RandomizationError(
                        f"Subject does not exist. Got {self.subject_identifier}"
                    )
                else:
                    raise AlreadyRandomized(
                        "Subject already randomized. See RegisteredSubject. "
                        f"Got {obj.subject_identifier} "
                        f"SID={obj.sid}",
                        code=get_registered_subject_model_cls()._meta.label_lower,
                    )
        return self._registered_subject

    @classmethod
    def get_extra_list_display(cls):
        """Returns a list of tuples of (pos, field name) for ModelAdmin."""
        return []

    @classmethod
    def get_extra_list_filter(cls):
        """Returns a list of tuples of (pos, field name) for ModelAdmin."""
        return cls.get_extra_list_display()

    @classmethod
    def verify_list(cls):
        randomization_list_verifier = RandomizationListVerifier(randomizer_name=cls.name)
        return randomization_list_verifier.messages

    @classmethod
    def import_list(cls, **kwargs):
        importer = cls.importer_cls(randomizer_cls=cls, **kwargs)
        importer.import_list()



import csv
import sys
from pprint import pprint
from uuid import uuid4

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style
from tqdm import tqdm

style = color_style()


class RandomizationListImportError(Exception):
    pass


class RandomizationListAlreadyImported(Exception):
    pass


class RandomizationListImporter:
    """Imports upon instantiation a formatted randomization CSV file
    into model RandomizationList.

    default CSV file is the projects randomization_list.csv

    name: name of randomizer, e.g. "default"

    To import SIDS from CSV for the first time:

        from edc_randomization.randomization_list_importer import RandomizationListImporter

        RandomizationListImporter(name='default', add=False, dryrun=False)

        # note: if this is not the first time you will get:
        # RandomizationListImportError: Not importing CSV.
        # edc_randomization.randomizationlist model is not empty!

    To add additional sids from CSV without touching existing model instances:

        from edc_randomization.randomization_list_importer import RandomizationListImporter

        RandomizationListImporter(name='default', add=True, dryrun=False)


    Format:
        sid,assignment,site_name, orig_site, orig_allocation, orig_desc
        1,single_dose,gaborone
        2,two_doses,gaborone
        ...
    """

    default_fieldnames = ["sid", "assignment", "site_name"]

    def __init__(
        self,
        randomizer_cls=None,
        verbose: bool = None,
        overwrite: bool = None,
        add: bool = None,
        dryrun: bool = None,
        username: str = None,
        revision: str = None,
        sid_count_for_tests: int = None,
    ):
        self.sid_count = 0
        self.randomizer_cls = randomizer_cls
        self.add = add
        self.overwrite = overwrite
        self.verbose = True if verbose is None else verbose
        self.dryrun = dryrun
        self.revision = revision
        self.user = username
        self.sid_count_for_tests = sid_count_for_tests

        if self.dryrun:
            sys.stdout.write(
                style.MIGRATE_HEADING("\n ->> Dry run. No changes will be made.\n")
            )
        if not self.get_site_names():
            raise RandomizationListImportError(
                "No sites have been imported. See sites module and ."
                'method "add_or_update_django_sites".'
            )
        if self.verbose and add:
            count = self.randomizer_cls.model_cls().objects.all().count()
            sys.stdout.write(
                style.SUCCESS(f"(*) Randolist model has {count} SIDs (count before import).\n")
            )

    def import_list(self):
        self._raise_on_invalid_header()
        self._raise_on_already_imported()
        self._raise_on_duplicates()
        self._import_to_model()
        self._summarize_results()
        sys.stdout.write(
            style.SUCCESS(
                f"(*) Loaded randomizer {self.randomizer_cls}.\n"
                f"    -  Name: {self.randomizer_cls.name}\n"
                f"    -  Assignments: {self.randomizer_cls.assignment_map}\n"
                f"    -  Blinded trial:  {self.randomizer_cls.is_blinded_trial}\n"
                f"    -  CSV file:  {self.randomizer_cls.filename}\n"
                f"    -  Model: {self.randomizer_cls.model}\n"
                f"    -  Path: {self.randomizer_cls.randomization_list_path}\n"
            )
        )
        return self.randomizer_cls.get_randomization_list_fullpath()

    def _summarize_results(self):
        if self.verbose:
            count = self.randomizer_cls.model_cls().objects.all().count()
            path = self.randomizer_cls.get_randomization_list_fullpath()
            sys.stdout.write(
                style.SUCCESS(
                    f"(*) Imported {count} SIDs for randomizer "
                    f"`{self.randomizer_cls.name}` into model "
                    f"`{self.randomizer_cls.model_cls()._meta.label_lower}` \n"
                    f"    from {path} (count after import).\n"
                )
            )
        if not self.randomizer_cls.get_randomization_list_fullpath():
            raise RandomizationListImportError("No randomization list to imported!")

    def _raise_on_invalid_header(self):
        with open(self.randomizer_cls.get_randomization_list_fullpath(), "r") as csvfile:
            reader = csv.DictReader(csvfile)
            for index, row in enumerate(reader):
                if index == 0:
                    for fieldname in self.default_fieldnames:
                        if fieldname not in row:
                            raise RandomizationListImportError(
                                "Invalid header. Missing column " f"`{fieldname}`. Got {row}"
                            )
                elif index == 1:
                    if self.dryrun:
                        row_as_dict = {k: v for k, v in row.items()}
                        print(" -->  First row:")
                        print(f" -->  {list(row_as_dict.keys())}")
                        print(f" -->  {list(row_as_dict.values())}")
                        obj = self.randomizer_cls.model_cls()(**self.get_import_options(row))
                        pprint(obj.__dict__)
                else:
                    break

    def _raise_on_already_imported(self):
        if not self.dryrun:
            if self.overwrite:
                self.randomizer_cls.model_cls().objects.all().delete()
            if self.randomizer_cls.model_cls().objects.all().count() > 0 and not self.add:
                raise RandomizationListAlreadyImported(
                    f"Not importing CSV. "
                    f"{self.randomizer_cls.model_cls()._meta.label_lower} model is not empty!"
                )

    def _raise_on_duplicates(self):
        with open(self.randomizer_cls.get_randomization_list_fullpath(), "r") as csvfile:
            reader = csv.DictReader(csvfile)
            sids = [row["sid"] for row in reader]
        if len(sids) != len(list(set(sids))):
            raise RandomizationListImportError("Invalid file. Detected duplicate SIDs")
        self.sid_count = len(sids)

    def _import_to_model(self):
        """Imports a CSV to populate the "rando" model"""
        objs = []
        with open(self.randomizer_cls.get_randomization_list_fullpath(), "r") as csvfile:
            reader = csv.DictReader(csvfile)
            if self.sid_count_for_tests:
                sys.stdout.write(
                    style.WARNING(
                        "\nNote: Importing a `subset` of the randomization list for tests\n"
                    )
                )
            sid_count = self.sid_count_for_tests or self.sid_count
            for row in tqdm(reader, total=sid_count):
                row = {k: v.strip() for k, v in row.items()}
                try:
                    self.randomizer_cls.model_cls().objects.get(sid=row["sid"])
                except ObjectDoesNotExist:
                    opts = self.get_import_options(row)
                    opts.update(self.get_extra_import_options(row))
                    if self.user:
                        opts.update(user_created=self.user)
                    if self.revision:
                        opts.update(revision=self.revision)
                    obj = self.randomizer_cls.model_cls()(**opts)
                    objs.append(obj)
                if len(objs) == sid_count:
                    break
            if not self.dryrun:
                sys.stdout.write(
                    style.SUCCESS(
                        f"\n    -  bulk creating {self.sid_count_for_tests or self.sid_count} "
                        "model instances ...\r"
                    )
                )
                self.randomizer_cls.model_cls().objects.bulk_create(objs)
                sys.stdout.write(
                    style.SUCCESS(
                        f"    -  bulk creating {self.sid_count_for_tests or self.sid_count} "
                        "model instances ... done\n"
                    )
                )
                rec_count = self.randomizer_cls.model_cls().objects.all().count()
                if not sid_count == rec_count:
                    raise RandomizationListImportError(
                        "Incorrect record count on import. "
                        f"Expected {sid_count}. Got {rec_count}."
                    )
                sys.stdout.write(
                    style.SUCCESS(
                        "    Important: You may wish to run the randomization list "
                        "verifier before going LIVE on production systems."
                    )
                )

            else:
                sys.stdout.write(
                    style.MIGRATE_HEADING(
                        "\n ->> this is a dry run. No changes were saved. **\n"
                    )
                )

    def get_import_options(self, row):
        return dict(
            id=uuid4(),
            sid=row["sid"],
            assignment=self.randomizer_cls.get_assignment(row),
            allocation=str(self.randomizer_cls.get_allocation(row)),
            randomizer_name=self.randomizer_cls.name,
            site_name=self.validate_site_name(row),
            **self.get_extra_import_options(row),
        )

    def get_extra_import_options(self, row):
        return {}

    @staticmethod
    def get_site_names():
        """A dict of site names for the target randomizer.

        Default: All sites"""
        from django.contrib.sites.models import Site

        sites = {obj.name: obj.name for obj in Site.objects.all()}
        if not sites:
            raise RandomizationListImportError(
                "No sites have been imported. See sites module and ."
                'method "add_or_update_django_sites".'
            )
        return sites

    def validate_site_name(self, row):
        """Returns the site name or raises"""
        try:
            site_name = self.get_site_names()[row["site_name"]]
        except KeyError:
            raise RandomizationListImportError(
                f"Invalid site. Got {row['site_name']}. "
                f"Expected one of {self.get_site_names().keys()}"
            )
        return site_name



from random import shuffle
from tempfile import mkdtemp

from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.test import TestCase
from django.test.utils import override_settings
from edc_constants.constants import FEMALE
from edc_registration.models import RegisteredSubject
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from multisite import SiteID

from edc_randomization.constants import ACTIVE
from edc_randomization.models import RandomizationList
from edc_randomization.randomization_list_importer import (
    RandomizationListAlreadyImported,
    RandomizationListImporter,
)
from edc_randomization.randomization_list_verifier import (
    RandomizationListError,
    RandomizationListVerifier,
)
from edc_randomization.randomizer import (
    AllocationError,
    AlreadyRandomized,
    InvalidAssignment,
    RandomizationError,
    Randomizer,
)
from edc_randomization.site_randomizers import NotRegistered, site_randomizers
from edc_randomization.utils import (
    RandomizationListExporterError,
    export_randomization_list,
)

from ..make_test_list import make_test_list
from ..models import SubjectConsent
from ..randomizers import MyRandomizer

fqdn = "example.clinicedc.org"
all_sites = (
    SingleSite(10, "site_one", title="One", country="uganda", country_code="ug", fqdn=fqdn),
    SingleSite(20, "site_two", title="Two", country="uganda", country_code="ug", fqdn=fqdn),
    SingleSite(
        30, "site_three", title="Three", country="uganda", country_code="ug", fqdn=fqdn
    ),
    SingleSite(40, "site_four", title="Four", country="uganda", country_code="ug", fqdn=fqdn),
    SingleSite(50, "site_five", title="Five", country="uganda", country_code="ug", fqdn=fqdn),
)


@override_settings(
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=True,
)
class TestRandomizer(TestCase):
    import_randomization_list = False
    site_names = [x.name for x in all_sites]

    def setUp(self):
        super().setUp()
        add_or_update_django_sites(sites=all_sites)
        site_randomizers._registry = {}
        site_randomizers.register(Randomizer)

    def populate_list(
        self, randomizer_name=None, site_names=None, per_site=None, overwrite_site=None
    ):
        randomizer = site_randomizers.get(randomizer_name)
        make_test_list(
            full_path=randomizer.get_randomization_list_fullpath(),
            site_names=site_names or self.site_names,
            per_site=per_site,
        )
        randomizer.import_list(overwrite=True)
        if overwrite_site:
            site = Site.objects.get_current()
            randomizer.model_cls().objects.update(site_name=site.name)

    @override_settings(SITE_ID=SiteID(40))
    def test_(self):
        randomizer = site_randomizers.get("default")
        randomizer.import_list()

    @override_settings(SITE_ID=SiteID(40))
    def test_with_consent_insufficient_data(self):
        randomizer = site_randomizers.get("default")
        randomizer.import_list()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        self.assertRaises(
            RandomizationError,
            Randomizer,
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=None,
        )

    @override_settings(SITE_ID=SiteID(40))
    def test_with_consent(self):
        randomizer = site_randomizers.get("default")
        randomizer.import_list()
        site = Site.objects.get_current()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", site=site, user_created="erikvw"
        )
        try:
            Randomizer(
                subject_identifier=subject_consent.subject_identifier,
                report_datetime=subject_consent.consent_datetime,
                site=subject_consent.site,
                user=subject_consent.user_created,
            )
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {str(e)}.")

    @override_settings(SITE_ID=SiteID(40))
    def test_with_gender_and_consent(self):
        class RandomizerWithGender(Randomizer):
            def __init__(self, gender=None, **kwargs):
                self.gender = gender
                super().__init__(**kwargs)

            @property
            def extra_required_attrs(self):
                return dict(gender=self.gender)

        randomizer = site_randomizers.get("default")
        randomizer.import_list()
        site = Site.objects.get_current()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", site=site, gender=FEMALE, user_created="erikvw"
        )
        try:
            RandomizerWithGender(
                subject_identifier=subject_consent.subject_identifier,
                report_datetime=subject_consent.consent_datetime,
                site=subject_consent.site,
                gender=FEMALE,
                user=subject_consent.user_created,
            )
        except Exception as e:
            self.fail(f"Exception unexpectedly raised. Got {str(e)}.")

    @override_settings(SITE_ID=SiteID(40))
    def test_with_list_selects_first(self):
        self.populate_list(randomizer_name="default", overwrite_site=True)
        first_obj = RandomizationList.objects.all().first()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        rando = Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        self.assertEqual(rando.sid, first_obj.sid)

    @override_settings(SITE_ID=SiteID(40))
    def test_updates_registered_subject(self):
        self.populate_list(randomizer_name="default", overwrite_site=True)
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        first_obj = RandomizationList.objects.all().first()
        rs = RegisteredSubject.objects.get(subject_identifier="12345")
        self.assertEqual(rs.subject_identifier, first_obj.subject_identifier)
        self.assertEqual(rs.sid, str(first_obj.sid))
        self.assertEqual(rs.randomization_datetime, first_obj.allocated_datetime)
        self.assertEqual(rs.randomization_list_model, first_obj._meta.label_lower)

    @override_settings(SITE_ID=SiteID(40))
    def test_updates_list_obj_as_allocated(self):
        self.populate_list(randomizer_name="default", overwrite_site=True)
        RandomizationList.objects.all().first()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        first_obj = RandomizationList.objects.all().first()
        self.assertEqual(first_obj.subject_identifier, "12345")
        self.assertTrue(first_obj.allocated)
        self.assertIsNotNone(first_obj.allocated_user)
        self.assertEqual(first_obj.allocated_user, subject_consent.user_created)
        self.assertEqual(first_obj.allocated_datetime, subject_consent.consent_datetime)
        self.assertGreater(first_obj.modified, subject_consent.created)

    @override_settings(SITE_ID=SiteID(40))
    def test_cannot_rerandomize(self):
        self.populate_list(randomizer_name="default", overwrite_site=True)
        first_obj = RandomizationList.objects.all().first()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        rando = Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        self.assertEqual(rando.sid, first_obj.sid)
        self.assertRaises(
            AlreadyRandomized,
            Randomizer,
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )

    @override_settings(SITE_ID=SiteID(40))
    def test_error_condition1(self):
        """Assert raises if RegisteredSubject not updated correctly."""
        self.populate_list(randomizer_name="default", overwrite_site=True)
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        rando = Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        rando.registered_subject.sid = None
        rando.registered_subject.save()
        with self.assertRaises(AlreadyRandomized) as cm:
            Randomizer(
                subject_identifier=subject_consent.subject_identifier,
                report_datetime=subject_consent.consent_datetime,
                site=subject_consent.site,
                user=subject_consent.user_created,
            )
        self.assertEqual(cm.exception.code, "edc_randomization.randomizationlist")

    @override_settings(SITE_ID=SiteID(40))
    def test_error_condition2(self):
        """Assert raises if RandomizationList not updated correctly."""
        self.populate_list(randomizer_name="default", overwrite_site=True)
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", user_created="erikvw"
        )
        rando = Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        rando.registered_subject.sid = None
        rando.registered_subject.save()
        with self.assertRaises(AlreadyRandomized) as cm:
            Randomizer(
                subject_identifier=subject_consent.subject_identifier,
                report_datetime=subject_consent.consent_datetime,
                site=subject_consent.site,
                user=subject_consent.user_created,
            )
        self.assertEqual(cm.exception.code, "edc_randomization.randomizationlist")

    def test_error_condition3(self):
        """Assert raises if RandomizationList not updated correctly."""
        self.populate_list(randomizer_name="default", overwrite_site=True)
        site = Site.objects.get_current()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", site=site, user_created="erikvw"
        )
        Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        RandomizationList.objects.update(subject_identifier=None)
        with self.assertRaises(AlreadyRandomized) as cm:
            Randomizer(
                subject_identifier=subject_consent.subject_identifier,
                report_datetime=subject_consent.consent_datetime,
                site=subject_consent.site,
                user=subject_consent.user_created,
            )
        self.assertEqual(cm.exception.code, "edc_registration.registeredsubject")

    def test_subject_does_not_exist(self):
        self.populate_list(randomizer_name="default", overwrite_site=True)
        site = Site.objects.get_current()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", site=site, user_created="erikvw"
        )
        RegisteredSubject.objects.all().delete()
        self.assertRaises(
            RandomizationError,
            Randomizer,
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_modified,
        )

    def test_str(self):
        self.populate_list(randomizer_name="default", overwrite_site=True)
        site = Site.objects.get_current()
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345", site=site, user_created="erikvw"
        )
        Randomizer(
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_created,
        )
        obj = RandomizationList.objects.all().first()
        self.assertTrue(str(obj))

    @override_settings(SITE_ID=SiteID(40))
    def test_for_sites(self):
        """Assert that allocates by site correctly."""

        site = None
        site_randomizers._registry = {}
        site_randomizers.register(MyRandomizer)

        model_cls = MyRandomizer.model_cls()
        model_cls.objects.all().delete()
        self.populate_list(
            randomizer_name=MyRandomizer.name, site_names=self.site_names, per_site=5
        )
        site_names = [obj.site_name for obj in model_cls.objects.all()]
        shuffle(site_names)
        self.assertEqual(len(site_names), len(self.site_names * 5))
        # consent and randomize 5 for each site
        for index, site_name in enumerate(site_names):
            site = Site.objects.get(name=site_name)
            subject_consent = SubjectConsent.objects.create(
                subject_identifier=f"12345{index}", site=site, user_created="erikvw"
            )
            MyRandomizer(
                subject_identifier=subject_consent.subject_identifier,
                report_datetime=subject_consent.consent_datetime,
                site=subject_consent.site,
                user=subject_consent.user_created,
            )
        # assert consented subjects were allocated SIDs in the
        # correct order per site.
        for site_name in site_names:
            randomized_subjects = [
                (obj.subject_identifier, str(obj.sid))
                for obj in model_cls.objects.filter(
                    allocated_site__name=site_name, subject_identifier__isnull=False
                ).order_by("sid")
            ]
            for index, obj in enumerate(
                SubjectConsent.objects.filter(site__name=site_name).order_by(
                    "consent_datetime"
                )
            ):
                rs = RegisteredSubject.objects.get(subject_identifier=obj.subject_identifier)
                self.assertEqual(obj.subject_identifier, randomized_subjects[index][0])
                self.assertEqual(rs.sid, randomized_subjects[index][1])

        # clear out any unallocated
        model_cls.objects.filter(subject_identifier__isnull=True).delete()

        # assert raises on next attempt to randomize
        subject_consent = SubjectConsent.objects.create(
            subject_identifier="ABCDEF",
            site=site,
            user_created="erikvw",
            user_modified="erikvw",
        )
        self.assertRaises(
            AllocationError,
            MyRandomizer,
            subject_identifier=subject_consent.subject_identifier,
            report_datetime=subject_consent.consent_datetime,
            site=subject_consent.site,
            user=subject_consent.user_modified,
        )

    @override_settings(SITE_ID=SiteID(40))
    def test_not_loaded(self):
        try:
            RandomizationListVerifier(randomizer_name=Randomizer.name)
        except RandomizationListError as e:
            self.assertIn("Randomization list has not been loaded", str(e))
        else:
            self.fail("RandomizationListError unexpectedly NOT raised")

    @override_settings(SITE_ID=SiteID(40))
    def test_cannot_overwrite(self):
        site_randomizers._registry = {}
        site_randomizers.register(MyRandomizer)
        make_test_list(
            full_path=MyRandomizer.get_randomization_list_fullpath(),
            site_names=self.site_names,
            count=5,
        )
        randomizer = site_randomizers.get(MyRandomizer.name)
        randomizer.import_list()
        self.assertRaises(RandomizationListAlreadyImported, randomizer.import_list)

    @override_settings(SITE_ID=SiteID(40))
    def test_can_overwrite_explicit(self):
        site_randomizers._registry = {}
        site_randomizers.register(MyRandomizer)
        make_test_list(
            full_path=MyRandomizer.get_randomization_list_fullpath(),
            site_names=self.site_names,
            count=5,
        )
        randomizer = site_randomizers.get(MyRandomizer.name)
        try:
            randomizer.import_list(overwrite=True)
        except RandomizationListAlreadyImported:
            self.fail("RandomizationListImportError unexpectedly raised")

    @override_settings(SITE_ID=SiteID(40))
    def test_invalid_assignment(self):
        site_randomizers._registry = {}
        site_randomizers.register(MyRandomizer)

        MyRandomizer.get_randomization_list_fullpath()
        make_test_list(
            full_path=MyRandomizer.get_randomization_list_fullpath(),
            site_names=self.site_names,
            # change to a different assignments
            assignments=[100, 101],
            count=5,
        )
        self.assertRaises(InvalidAssignment, MyRandomizer.import_list)

    @override_settings(SITE_ID=SiteID(40))
    def test_invalid_sid(self):
        self.populate_list(randomizer_name="default")
        # change to a different starting SID
        obj = RandomizationList.objects.all().order_by("sid").first()
        obj.sid = 99999
        obj.save()

        with self.assertRaises(RandomizationListError) as cm:
            RandomizationListVerifier(randomizer_name=Randomizer.name)
        self.assertIn("Randomization list has invalid SIDs", str(cm.exception))

    @override_settings(SITE_ID=SiteID(40))
    def test_invalid_count(self):
        site = Site.objects.get_current()
        # change number of SIDs in DB
        self.populate_list(randomizer_name="default")
        RandomizationList.objects.create(sid=100, assignment=ACTIVE, site_name=site.name)
        self.assertEqual(RandomizationList.objects.all().count(), 51)
        with self.assertRaises(RandomizationListError) as cm:
            RandomizationListVerifier(randomizer_name=Randomizer.name)
        self.assertIn("Randomization list count is off", str(cm.exception))

    @override_settings(SITE_ID=SiteID(40))
    def test_get_randomizer_cls(self):
        site_randomizers._registry = {}
        self.assertRaises(NotRegistered, site_randomizers.get, MyRandomizer.name)
        site_randomizers.register(MyRandomizer)
        try:
            site_randomizers.get(MyRandomizer.name)
        except NotRegistered:
            self.fail("NotRegistered unexpectedly raised")

    @override_settings(SITE_ID=SiteID(40))
    def test_randomization_list_importer(self):
        randomizer_cls = site_randomizers.get("default")

        make_test_list(
            full_path=randomizer_cls.get_randomization_list_fullpath(),
            site_names=self.site_names,
        )

        importer = RandomizationListImporter(randomizer_cls, dryrun=True, verbose=True)
        importer.import_list()
        self.assertEqual(randomizer_cls.model_cls().objects.all().count(), 0)

        importer = RandomizationListImporter(randomizer_cls, verbose=True)
        importer.import_list()
        self.assertGreater(randomizer_cls.model_cls().objects.all().count(), 0)

    @override_settings(SITE_ID=SiteID(40), EXPORT_FOLDER=mkdtemp())
    def test_randomization_list_exporter(self):
        user = get_user_model().objects.create(
            username="me", is_superuser=False, is_staff=True
        )
        randomizer_cls = site_randomizers.get("default")
        make_test_list(
            full_path=randomizer_cls.get_randomization_list_fullpath(),
            site_names=self.site_names,
        )
        importer = RandomizationListImporter(randomizer_cls, verbose=True)
        importer.import_list()
        self.assertRaises(RandomizationListExporterError, export_randomization_list, "default")
        self.assertRaises(
            RandomizationListExporterError,
            export_randomization_list,
            "default",
            username=user.username,
        )
        user = get_user_model().objects.create(
            username="you", is_superuser=True, is_staff=True
        )
        path = export_randomization_list("default", username=user.username)
        with open(path) as f:
            n = 0
            for line in f:
                n += 1
                if "str" in line:
                    break
        self.assertEqual(n, 51)



from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.utils.safestring import mark_safe

from .auth_objects import RANDO_UNBLINDED


def is_blinded_trial():
    return getattr(settings, "EDC_RANDOMIZATION_BLINDED_TRIAL", True)


def is_blinded_user(username):
    if is_blinded_trial():
        _is_blinded_user = True
        unblinded_users = getattr(settings, "EDC_RANDOMIZATION_UNBLINDED_USERS", [])
        try:
            user = get_user_model().objects.get(
                username=username, is_staff=True, is_active=True
            )
        except ObjectDoesNotExist:
            pass
        else:
            if user.username in unblinded_users:
                _is_blinded_user = False
    else:
        _is_blinded_user = False
    return _is_blinded_user


def raise_if_prohibited_from_unblinded_rando_group(username, groups):
    """A user form validation to prevent adding an unlisted
    user to the RANDO_UNBLINDED group.

    See also edc_auth's UserForm.
    """
    if RANDO_UNBLINDED in [grp.name for grp in groups] and is_blinded_user(username):
        raise forms.ValidationError(
            {
                "groups": mark_safe(
                    "This user is not unblinded and may not added "
                    "to the <U>RANDO_UNBLINDED</U> group."
                )
            }
        )



import os
from typing import Optional

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from edc_pdutils.model_to_dataframe import ModelToDataframe
from edc_utils import get_utcnow

from ..site_randomizers import site_randomizers


class RandomizationListExporterError(Exception):
    pass


def export_randomization_list(
    randomizer_name: str, path: Optional[str] = None, username: Optional[str] = None
):
    randomizer_cls = site_randomizers.get(randomizer_name)

    try:
        user = get_user_model().objects.get(username=username)
    except ObjectDoesNotExist:
        raise RandomizationListExporterError(f"User `{username}` does not exist")
    if not user.has_perm(randomizer_cls.model_cls()._meta.label_lower.replace(".", ".view_")):
        raise RandomizationListExporterError(
            f"User `{username}` does not have "
            f"permission to view '{randomizer_cls.model_cls()._meta.label_lower}'"
        )
    path = path or settings.EXPORT_FOLDER
    timestamp = get_utcnow().strftime("%Y%m%d%H%M")
    filename = os.path.expanduser(
        f"~/{settings.APP_NAME}_{randomizer_cls.name}_"
        f"randomizationlist_exported_{timestamp}.csv"
    )
    filename = os.path.join(path, filename)

    df = ModelToDataframe(
        model=randomizer_cls.model_cls()._meta.label_lower, decrypt=True, drop_sys_columns=True
    )
    opts = dict(
        path_or_buf=filename,
        encoding="utf-8",
        index=0,
        sep="|",
    )
    df.dataframe.to_csv(**opts)
    print(filename)
    return filename



# Generated by Django 2.2.9 on 2020-02-02 20:01

import django_crypto_fields.fields.encrypted_char_field
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_randomization", "0004_auto_20191108_0752"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalrandomizationlist",
            name="assignment",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True, help_text=" (Encryption: RSA local)", max_length=71
            ),
        ),
        migrations.AlterField(
            model_name="randomizationlist",
            name="assignment",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True, help_text=" (Encryption: RSA local)", max_length=71
            ),
        ),
    ]



# Generated by Django 2.2.9 on 2020-02-13 17:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_randomization", "0005_auto_20200202_2301"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalrandomizationlist",
            name="randomizer_name",
            field=models.CharField(default="default", max_length=50),
        ),
        migrations.AddField(
            model_name="randomizationlist",
            name="randomizer_name",
            field=models.CharField(default="default", max_length=50),
        ),
    ]



import copy
import sys

from django.apps import apps as django_apps
from django.utils.module_loading import import_module, module_has_submodule


class RegistryNotLoaded(Exception):
    pass


class NotRegistered(Exception):
    pass


class AlreadyRegistered(Exception):
    pass


class SiteRandomizerError(Exception):
    pass


class SiteRandomizers:
    """Main controller of :class:`SiteRandomizers` objects."""

    def __init__(self):
        self._registry = {}
        self.loaded = False

    @property
    def registry(self):
        if not self.loaded:
            raise RegistryNotLoaded(
                "Registry not loaded. Is AppConfig for 'edc_randomization' "
                "declared in settings?."
            )
        return self._registry

    def register(self, randomizer_cls):
        self.loaded = True
        if str(randomizer_cls.name) not in self.registry:
            self.registry.update({str(randomizer_cls.name): randomizer_cls})
        else:
            raise AlreadyRegistered(
                f"Randomizer class for `{randomizer_cls}` is already registered. "
                f"Got name=`{randomizer_cls.name}`. See also `settings` attribute "
                "`EDC_RANDOMIZATION_REGISTER_DEFAULT_RANDOMIZER`."
            )

    def get(self, name):
        try:
            return self._registry[str(name)]
        except KeyError:
            raise NotRegistered(
                f"A Randomizer class by this name is not registered. "
                f"Expected one of {list(self._registry.keys())}. "
                f"Got '{name}'. See site_randomizer."
            )

    def get_by_model(self, model=None):
        """Returns the randomizer class for this model label_lower.

        Default: `edc_randomization.randomizationlist`.
        """
        model = model or "edc_randomization.randomizationlist"
        for randomizer_cls in self._registry.values():
            if randomizer_cls.model == model:
                return randomizer_cls
        return None

    def get_as_choices(self):
        choices = []
        for randomizer_cls in self._registry.values():
            choices.append((randomizer_cls.name, randomizer_cls.name))
        return tuple(choices)

    def randomize(
        self,
        name,
        subject_identifier=None,
        report_datetime=None,
        site=None,
        user=None,
        **kwargs,
    ):
        randomizer_cls = self.get(name)
        return randomizer_cls(
            subject_identifier=subject_identifier,
            report_datetime=report_datetime,
            site=site,
            user=user,
            **kwargs,
        )

    def autodiscover(self, module_name=None, apps=None, verbose=None):
        """Autodiscovers classes in the randomizers.py file of
        any INSTALLED_APP.
        """
        if not self.loaded:
            self.loaded = True
            module_name = module_name or "randomizers"
            verbose = True if verbose is None else verbose
            if verbose:
                sys.stdout.write(f" * checking site for module '{module_name}' ...\n")
            for app in apps or django_apps.app_configs:
                try:
                    mod = import_module(app)
                    try:
                        before_import_registry = copy.copy(site_randomizers._registry)
                        import_module(f"{app}.{module_name}")
                        if verbose:
                            sys.stdout.write(" * registered randomizer from " f"'{app}'\n")
                    except Exception as e:
                        if f"No module named '{app}.{module_name}'" not in str(e):
                            raise
                        site_randomizers._registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise
                except ModuleNotFoundError:
                    pass


site_randomizers = SiteRandomizers()



from django.db import models
from edc_model.models import BaseUuidModel
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow

from edc_randomization.models import RandomizationListModelMixin

from .randomizers import MyRandomizer


class SubjectConsent(UpdatesOrCreatesRegistrationModelMixin, SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=25)

    initials = models.CharField(max_length=25)

    consent_datetime = models.DateTimeField(default=get_utcnow)

    gender = models.CharField(max_length=25, null=True)


class MyRandomizationList(RandomizationListModelMixin, BaseUuidModel):
    randomizer_cls = MyRandomizer

    class Meta(RandomizationListModelMixin.Meta):
        pass



# -*- coding: utf-8 -*-
import os
from os.path import abspath, dirname, join, normpath

from setuptools import find_packages, setup

with open(join(dirname(__file__), "README.rst")) as readme:
    README = readme.read()

with open(join(dirname(__file__), "VERSION")) as f:
    VERSION = f.read()

# allow setup.py to be run from any path
os.chdir(normpath(join(abspath(__file__), os.pardir)))

setup(
    name="edc-randomization",
    version=VERSION,
    author="Erik van Widenfelt",
    author_email="ew2789@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    url="http://github.com/clinicedc/edc-randomization",
    license="GPL license, see LICENSE",
    long_description_content_type="text/x-rst",
    description="Randomization classes for clinicedc/edc projects.",
    long_description=README,
    zip_safe=False,
    keywords="django base models fields forms admin",
    install_requires=[],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    python_requires=">=3.8",
)



import csv
import os
import uuid

from django.apps import apps as django_apps
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist


class AllocationError(Exception):
    pass


def get_assignment(allocation):
    if allocation == 1:
        return "placebo"
    elif allocation == 2:
        return "active"
    else:
        raise AllocationError(f"Invalid allocation. Got {allocation}")


def get_allocation(assignment):
    if assignment == "placebo":
        return "1"
    elif assignment == "active":
        return "2"
    else:
        raise AllocationError(f"Invalid assignment. Got {assignment}")


def new_rows(allocations=None, site_name=None, allocation_map=None):
    rows = []
    for index, allocation in enumerate(allocations, 1):
        rows.append(
            dict(
                sid=index,
                allocation=str(allocation),
                assignment=allocation_map[allocation],
                site_name=site_name,
                site_id=Site.objects.get(name=site_name).id,
            )
        )
    return rows


def import_allocations(
    model=None, allocations=None, site_name=None, allocation_map=None, dry_run=None
):
    rows = new_rows(
        allocations=allocations, site_name=site_name, allocation_map=allocation_map
    )
    randomizationlist_model_cls = django_apps.get_model(model)
    last_sid = randomizationlist_model_cls.objects.all().order_by("sid").last().sid
    objs = []
    for row in rows:
        row = {k: v.strip() for k, v in row.items()}
        try:
            obj = randomizationlist_model_cls.objects.get(sid=last_sid + row["sid"])
        except ObjectDoesNotExist:
            obj = randomizationlist_model_cls(
                id=uuid.uuid4(),
                sid=last_sid + row["sid"],
                assignment=row["assignment"],
                site_name=row["site"],
                allocation=get_allocation(row["assignment"]),
            )
            objs.append(obj)
    if dry_run:
        print(objs)
    else:
        randomizationlist_model_cls.objects.bulk_create(objs)


def import_additional_sids_from_file(
    filename=None, model=None, allocations=None, dry_run=None
):
    objs = []
    filename = filename or "~/rando_additional.txt"
    randomizationlist_model_cls = django_apps.get_model(model)
    last_sid = randomizationlist_model_cls.objects.all().order_by("sid").last().sid

    with open(os.path.join(os.path.expanduser(filename)), "r") as f:
        reader = csv.DictReader(f)
        for index, row in enumerate(reader, 1):
            row = {k: v.strip() for k, v in row.items()}
            try:
                obj = randomizationlist_model_cls.objects.get(sid=last_sid + index)
            except ObjectDoesNotExist:
                obj = randomizationlist_model_cls(
                    id=uuid.uuid4(),
                    sid=last_sid + index,
                    assignment=row["assignment"],
                    site_name=row["site"],
                    allocation=get_allocation(row["assignment"]),
                )
                objs.append(obj)
    if dry_run:
        print(objs)
    else:
        randomizationlist_model_cls.objects.bulk_create(objs)



import os
import sys
from collections import namedtuple

from django.conf import settings
from django.core.checks import Warning
from django.core.management import color_style

from .site_randomizers import site_randomizers

err = namedtuple("Err", "id cls")

error_configs = dict(randomization_list_check=err("edc_randomization.W001", Warning))

style = color_style()


def randomization_list_check(app_configs, **kwargs):
    sys.stdout.write(style.SQL_KEYWORD("randomization_list_check ... \r"))
    errors = []
    error = error_configs.get("randomization_list_check")
    for randomizer in site_randomizers.registry.values():
        if (
            "tox" not in sys.argv
            and "test" not in sys.argv
            and "runtests.py" not in sys.argv
            and "showmigrations" not in sys.argv
            and "makemigrations" not in sys.argv
            and "migrate" not in sys.argv
            and "shell" not in sys.argv
        ):
            error_msgs = randomizer.verify_list()
            for error_msg in error_msgs:
                errors.append(error.cls(error_msg, hint=None, obj=None, id=error.id))
        if not settings.DEBUG:
            if settings.ETC_DIR not in randomizer.get_randomization_list_fullpath():
                errors.append(
                    Warning(
                        "Insecure configuration. Randomization list file must be "
                        "stored in the etc folder. Got "
                        f"{randomizer.get_randomization_list_fullpath()}",
                        id="randomization_list_path",
                    )
                )
            if os.access(randomizer.get_randomization_list_fullpath(), os.W_OK):
                errors.append(
                    Warning(
                        "Insecure configuration. File is writeable by this user. "
                        f"Got {randomizer.get_randomization_list_fullpath()}",
                        id="randomization_list_path",
                    )
                )
    sys.stdout.write(style.SQL_KEYWORD("randomization_list_check ... done.\n"))
    return errors
